package lc.kra.cell;


import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import lc.kra.cell.gui.Provider;

/**
 * Wolfram Universe 
 *
 * Copyright 2011, 2013 Kristian Kraljic
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at

 * http://www.apache.org/licenses/LICENSE-2.0

 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * @author <a href=mailto:kristian@kra.lc>Kristian Kraljic</a>
 */
public class WolframUniverse extends JComponent implements Provider<WolframUniverse> {
	public enum Rule {
		RULE_1("Rule 1",new boolean[][] { {false,false,false} }),
		RULE_2("Rule 2",new boolean[][] { {false,false,true } }),
		RULE_3("Rule 3",new boolean[][] { {false,false,true },
		                                  {false,false,false} }),
		RULE_4("Rule 4",new boolean[][] { {false,true ,false} }),
		RULE_5("Rule 5",new boolean[][] { {false,true ,false},
	                                    {false,false,false} }),
	  RULE_110("Rule 110",new boolean[][] { {true ,true ,false},
										                      {true ,false,true },
										                      {false,true ,true }, 
										                      {false,true ,false}, 
										                      {false,false,true } }),
	  RULE_122("Rule 122",new boolean[][] { {true ,true ,false},
																					{true ,false,true },
																					{true ,false,false},
																					{false,true ,true },
																					{false,false,true } }),
    RULE_EXTENDED("Extended Rule",null);
		public final String name;
		public final boolean[][] rule;
		Rule(String name,boolean[][] rules) {
			this.name = name;
			this.rule = rules;
		}
		@Override public String toString() { return name; }
	}
	
	private static final long serialVersionUID = 1l;
	protected final static Map<Attribute,String> attributes;
	static {
		attributes = new HashMap<Attribute,String>();
		attributes.put(Attribute.NAME,"Wolfram Universe");
		attributes.put(Attribute.IDEA,"Stephen Wolfram");
		attributes.put(Attribute.CODE,"Kristian Kraljic");
		attributes.put(Attribute.COPYRIGHT,"(c) 2013 Kristian Kraljic");
		attributes.put(Attribute.DESCRIPTION,"An elementary cellular automaton is a one-dimensional cellular automaton where there are two possible states (labeled 0 and 1) and the rule to determine the state of a cell in the next generation depends only on the current state of the cell and its two immediate neighbors. As such it is one of the simplest possible models of computation.");
		attributes.put(Attribute.SOURCE,"http://en.wikipedia.org/wiki/Elementary_cellular_automaton");
	}
	
	public boolean[][] rule;
	public final int SIZE = 300, LIFESPAN = 1000;
	private static final int CELL = 3;
	
	private final boolean[] EXTINGUISHED = new boolean[SIZE];
	private Vector<boolean[]> universe;
	
	public WolframUniverse() {
		setBackground(Color.WHITE);
		extinguish(); //erase everything
		compute(); //big bang and first 'LIEFESPAN' years
	}
	
	public int age() { return universe.size(); }
	
	public void compute() { this.compute(null); }
	public void compute(boolean[] bigbang) {
		while(universe.size()<LIFESPAN)
			if(universe.isEmpty()) { //big bang
				if(bigbang==null||bigbang.length!=SIZE) {
					bigbang = new boolean[SIZE];
					for(int position=0;position<SIZE;position++)
						bigbang[position] = Math.random()<0.5;
					universe.add(bigbang);
				} else universe.add(bigbang);
			} else {
				boolean[] present = universe.lastElement();
				boolean[] future = new boolean[SIZE];
				for(int position=0;position<SIZE;position++)
					future[position] = computeCloseRangeEffect(present,position);
				universe.add(future);
				if(Arrays.equals(EXTINGUISHED,future))
					break;
			}
		this.repaint();
		this.setPreferredSize(new Dimension(SIZE*CELL,universe.size()*CELL));
	}
	
	public void extinguish() {
		universe = new Vector<boolean[]>();
	}
	
	private boolean computeCloseRangeEffect(boolean[] present,int position) {
		if(rule!=null) {
			for(boolean[] rule:this.rule) {
				boolean match = true;
				for(int range=0,element=position-(rule.length/2);range<rule.length;range++,element++)
					if(present[element<0?SIZE+element:(element>=SIZE?element-SIZE:element)]!=rule[range]) {
						match = false;
						break;
					}
				if(match)
					return true;
			}
			return false;
		} else {
			int count = 0;
			for(int range=0,element=position-2;range<5;range++,element++)
				if(present[element<0?SIZE+element:(element>=SIZE?element-SIZE:element)])
				  count++;
			return count==2||count==4;
		}
	}
	
	public void paintComponent(Graphics graphics) {
		super.paintComponent(graphics);
		graphics.setColor(getBackground());
		graphics.fillRect(0,0,getWidth(),getHeight());
		
		graphics.setColor(Color.BLACK);
		int top = 0;
		for(boolean[] history:universe) {
			int left = 0;
			for(boolean element:history) {
				if(element)
					graphics.fillRect(left,top,CELL,CELL);
				left+=CELL;
			}
			top += CELL;
		}
	}

	@Override public Map<Attribute,String> getAttributes() { return attributes; }
	@Override public WolframUniverse getVisualization() { return this; }
	@Override public Component getControls(WolframUniverse visualization) {
		JPanel controls = new JPanel(new FlowLayout(FlowLayout.LEFT,10,0));
		
		controls.add(new JLabel("Extinguish this universe and create another big bang:"));
		JButton single = new JButton("Singular");
		controls.add(single);
		single.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent actionevent) {
				boolean[] bigbang = new boolean[SIZE];
				bigbang[(int)(Math.random()*SIZE)] = true;
				extinguish(); compute(bigbang);
				getParent().revalidate();
			}
		});
		JButton random = new JButton("Chaos");
		controls.add(random);
		random.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent actionevent) {
				extinguish(); compute();
				getParent().revalidate();
			}
		});
		
		controls.add(new JLabel("using"));
		final JComboBox<Rule> rules = new JComboBox<Rule>(Rule.values());
		controls.add(rules);
		rules.setSelectedItem(Rule.RULE_EXTENDED);
		rules.addItemListener(new ItemListener() {
			@Override public void itemStateChanged(ItemEvent event) {
				rule = rules.getItemAt(rules.getSelectedIndex()).rule;
				boolean[] bigbang = universe.size()!=0?universe.get(0):null;
				extinguish(); compute(bigbang);
				getParent().revalidate();
			}
		});
		
		return controls;
	}

	@Override public void setUp() {
		compute();
		((JScrollPane)getParent().getParent()).setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
	}

	@Override public void tearDown() {
		extinguish();
		rule = null;
	}
	
	@Override public String toString() { return attributes.get(Attribute.NAME); }
}
